import sys
import math
#Session05:
import random

from OpenGL.GL  import *
from OpenGL.GLU import *


import pygame
from pygame.locals import *




class Light(object):
    enabled = False
    colors = [(1.,1.,1.,1.), (1.,0.5,0.5,1.),
            (0.5,1.,0.5,1.), (0.5,0.5,1.,.1,)]

    def __init__(self, light_id, position):
        self.light_id = light_id
        self.position = position
        self.current_color = 0 #current color index from the array

    def render(self):
        light_id = self.light_id
        color = Light.colors[self.current_color]
        glLightfv(light_id, GL_POSITION, self.position)
        glLightfv(light_id, GL_DIFFUSE, color)
        glLightfv(light_id, GL_CONSTANT_ATTENUATION, 0.1)
        glLightfv(light_id, GL_LINEAR_ATTENUATION, 0.05)

    def switch_color(self):
        self.current_color += 1  #next color
        self.current_color %= len(Light.colors) # wrap-arround

    def enable(self):
        if not Light.enabled:
            glEnable(GL_LIGHTING)
            Light.enabled = True
        glEnable(self.light_id)
 

#The sides of the cube are: the back face (0,1,2,3) the right face is (4,5,1,0)

# We arrange the vertices in counterclockwise, which will help in the optimization
# face culling which consists of drawing only the visible faces of the polygon

class Cube(object):
    sides = ((0,1,2,3), (3,2,7,6), (6,7,5,4),
             (4,5,1,0), (1,5,7,2), (4,0,3,6))

    edges = (
            (0,1),
            (0,3),
            (0,4),
            (2,1),
            (2,3),
            (2,7),
            (6,3),
            (6,4),
            (6,7),
            (5,1),
            (5,4),
            (5,7),
    )
    
    #save position/color/vertex coortinates
    def __init__(self, position, size, color):
        self.position = position
        self.color = color
        #lamba is used to define anonymous function
        #the lamda function is equivelant with:
        #def divide_by_2(i):
        # return i/2

        #map() apply a function to each item in an iterable (list/tupple) and return new iterable
        x,y,z = map(lambda i: i/2, size)

        self.vertices = (
            (x,-y,-z), (x,y,-z),
            (-x,y,-z), (-x, -y, -z),
            (x,-y,z), (x,y,z),
            (-x,-y, z), (-x, y,z)
        )

        
    #we add a render function for this object
    def render(self):
        glPushMatrix()
        #Move (translate) the current matrix by a certain amount along the x,y,z axes
        #glTranslatef(x,y,z). The x,y,z are the amounts that you want to translate the current
        #matrix along the x,y,z axis. The *self.position is used to unpack a tupple  or
        # list of three variables and pass them as seperate arguments
        glTranslatef(*self.position)

        glBegin(GL_QUADS) #treat each group of four vertices as independed quadrilateral
        # glMaterialfv(GL_FRONT, GL_DIFFUSE, self.color)
        # glMaterial(GL_FRONT, GL_SPECULAR, self.color)
        # glMaterial(GL_FRONT, GL_SHININESS, 10.0)
        for side in Cube.sides:            
            for v in side:
                #list of three floats that specify vertex
                #executed after glBegin and before glEnd
                #delimite the vertices
                glVertex3fv(self.vertices[v])
        glEnd()

        # glShadeModel(GL_SMOOTH)
        # #glBegin(GL_LINES)
        # glBegin(GL_POLYGON)
                
        # for edge in Cube.edges:
        #     glColor3fv((0.0,1.0,0.0))
        #     for vertex in edge:
        #         glVertex3fv(self.vertices[vertex])
        # glEnd()



        glPopMatrix()


class Sphere(object):
        slices=40
        stacks=40

        def __init__(self,radius,position,color):
            self.radius= radius
            self.position=position
            self.color = color
            #We have to create a quadratic object first
            #This is used to create basic 3D shapes.
            #This pointer is used to set various properties like orientation, position, size
            self.quadratic = gluNewQuadric()
        
        def render(self):
            glPushMatrix()
            glTranslatef(*self.position)
            glMaterialfv(GL_FRONT, GL_DIFFUSE, self.color)
            gluSphere(self.quadratic, self.radius, Sphere.slices, Sphere. stacks)
            glPopMatrix()




#Session05: we add the Block class which inherits from the Cube
class Block(Cube):
    color = (0, 0, 1. , 1.)
    speed = 0.01

    def __init__(self, position, size):
        #cube: position, size, color
        super().__init__(position, (size, 1, 1), Block.color)
        self.size=size

    #the update simply movs the block towards the player by updating its z coordinate
    def update(self,dt):
        x,y,z = self.position
        z += Block.speed * dt
        self.position = x, y, z








class App(object):
    def __init__(self, width=800, height=600):
        self.title= 'OpenGL Course Demo' 
        self.fps=60
        self.width= width
        self.height=height
        self.angle=0
        self.distance=20
        #Session05:
        self.game_over = False
        self.random_dt = 0
        self.blocks = []
        self.light = Light(GL_LIGHT0, (0,15,-25,1))
        self.player= Sphere(1, position=(0,0,0), color=(0,1,0,1))
        self.ground = Cube(position=(0,-1,-20), size=(16,1,60), color=(1,1,1,1))




        #Session05: we comment out everything
        #self.light=Light(GL_LIGHT0, (15,5,15,1))
        #Sphere(radius, position, color)
        #self.sphere1 = Sphere(2, (0,0,0),(1,1,1,1))
        #self.sphere2 = Sphere(1,(4,2,0), (1,0.4,0.4,1))
        #self.sphere3 = Sphere(0.5,(2,2,0), (1,0,0,1))
        #cube (position, size, color)
        #self.cube1 = Cube((5,4,0),(2,2,2),(0.678,0.847,0.902,1.0))
        #self.cube1 = Cube((-5,-4,0),(2,2,2),(0.0, 0.0, 1.0, 1.0))


    def start(self):
        pygame.init()
        pygame.display.set_mode((self.width, self.height), OPENGL | DOUBLEBUF)
        pygame.display.set_caption(self.title)

        #enable GL capabilities
        #=>depth comparions and update depth buffer
        glEnable(GL_DEPTH_TEST)
        #=>enables lighting
        glEnable(GL_LIGHTING)
        #=>enables light constant 0 (up to 8 can be used)
        glEnable(GL_LIGHT0)

        
        

       
        glEnable(GL_DEPTH_TEST)
        self.light.enable()


        #NOTE: with lighting, the colors do not use the glColor
        #but the lighting computation AND glMaterial
        #
        #if you like lighting with glColor then enable 
        #glEnable(GL_COLOR_MATERIAL)
        #glColor4f(r,g,b,a) #=> red, green, blue, alpha (0.0 zero intensity - 1.0 full intensity=brightness)

        #clear values, default value if nothing is rendered at this pixel
        glClearColor(.1,.1,.1,.1) 
        
        #set matrix mode for matrix operations. They are used in transformations
        glMatrixMode(GL_PROJECTION)
        aspect = self.width / self.height
        #generate perspetive projection matrix
        #view angle in y direction, aspect ratio for field of view,
        # , zNear distane from veiwer to the near plane, zfar distance from the viewer to the far plane
        gluPerspective(40.,aspect, 1., 40.)

        #set modelview mode, which is initial value, as current matrixmode
        glMatrixMode(GL_MODELVIEW)
        #specifies whether front--back-facing polygons use culling (are hidding)
        #CULLING is a mode used in many 3D libraries (even in direct3d)
        #
        #only renders the faces that are visible every time
        #Session05: enable culling
        glEnable(GL_CULL_FACE)
               
        clock=pygame.time.Clock()
        while True:
            dt= clock.tick(self.fps)
            self.process_input(dt)
            #Session05: changing the name to main_loop
            self.main_loop()

    #Session05: create the main loop 
    def main_loop(self):
        clock = pygame.time.Clock()
        while True:
            for event in pygame.event.get():
                if event.type == QUIT:
                    pygame.quit()
                    sys.exit()
                if not self.game_over:
                    self.display()
                    dt = clock.tick(self.fps)
                    for block in self.blocks:
                        block.update(dt)
                    self.clear_past_blocks()
                    self.add_random_block(dt)
                    self.check_collisions()
                    self.process_input(dt)

    #Session05:
    #compare the boundaries of the closed blocks with the extremes of the sphere.
    #the sphere is mall so if one the extremes is between right/left boundaries is a collision
    def check_collisions(self):
        #return to the blocks only these whose z position is between 0 and 1
        blocks = filter(lambda x: 0<x.position[2]<1, self.blocks)
        x=self.player.position[0]
        r=self.player.radius
        for block in blocks:
            x1= block.position[0]
            s=block.size / 2
            if x1-s < x-r < x1+s or x1-s < x +r < x1+s:
                self.game_over = True
                print("Game Over!")

    #Session05:
    def add_random_block(self, dt):
        self.random_dt +=dt
        if self.random_dt >= 800:
            r=random.random()
            if r < 0.1:
                self.random_dt=0
                self.generate_block(r)

    #Session05:
    def generate_block(self, r):
        size= 7 if r < 0.03 else 5
        offset = random.choice([-4,0,4])
        position=(offset, 0, -40)
        mynewblock=Block( position, size)
        self.blocks.append(mynewblock)

    #Session05:
    def clear_past_blocks(self):
        #extract the blocks with the given condition
        blocks = filter(lambda x: x.position[2] > 5, self.blocks)
        for block in blocks:
            self.blocks.remove(block)
            del block


    def process_input(self, dt):
        for event in pygame.event.get():
            if event.type == QUIT:
                self.quit()

            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.quit()
                if event.key == K_F1:
                    self.light.switch_color()

        pressed = pygame.key.get_pressed()
        #Session05:
        x, y, z = self.player.position
        
        if pressed[K_LEFT]:
            x -= 0.01 * dt

        if pressed[K_RIGHT]:
            x += 0.01 * dt

        #the player cannot move outside of the line
        x = max(min(x, 7), -7)
        self.player.position=(x,y,z)



    
    #Session05:
    def display(self):
      
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()
        gluLookAt(  0, 10, 10,
                    0, 0, -5,
                    0, 1, 0)
        self.light.render()
        for block in self.blocks:
            block.render()
        self.player.render()
        self.ground.render()
        pygame.display.flip()


    def quit(self):
        pygame.quit()
        sys.exit()





if __name__ == '__main__':
    app = App()
    app.start()
